from .autostepper import AutoStepper  # noqa: F401,F403
from .data import GrowingNCADataset  # noqa: F401
from .losses import *  # noqa: F401,F403
from .models import *  # noqa: F401,F403
from .paths import *  # noqa: F401,F403
from .prediction import *  # noqa: F401,F403
from .search import *  # noqa: F401,F403
from .training import *  # noqa: F401,F403
from .utils import *  # noqa: F401,F403
from .visualization import *  # noqa: F401,F403



from typing import Optional

import numpy as np
import torch


class Prediction:
    """
    Stores the result of an NCA prediction, including the number of steps it took.

    Sequences are typically stored by BasicNCAModel's "record" function, and are
    returned as a list of Prediction objects.
    """

    def __init__(self, model, steps: int, output_image: torch.Tensor):
        """
        Constructor is typically not called explicitly.
        Rather, the forward pass of BasicNCAModel (and its
        subclasses) is responsible for filling its attributes.

        :param model: Reference to model used for prediction.
        :type model: ncalab.BasicNCAModel
        :param steps: Number of steps taken for the prediction.
        :type steps: int
        :param output_image: Output image tensor.
        :type output_image: torch.Tensor
        """
        self.model = model
        self.steps = steps
        assert output_image.shape[1] == model.num_channels
        self.output_image = output_image
        self._output_array: Optional[np.ndarray] = None

    @property
    def image_channels(self) -> torch.Tensor:
        """
        Convenience property to access the image channels as a Tensor.

        :returns: BCWH Tensor
        :rtype: torch.Tensor
        """
        return self.output_image[:, : self.model.num_image_channels, :, :]

    @property
    def hidden_channels(self) -> torch.Tensor:
        """
        Convenience property to access the hidden channels as a Tensor.

        :returns: BCWH Tensor
        :rtype: torch.Tensor
        """
        return self.output_image[
            :,
            self.model.num_image_channels : self.model.num_hidden_channels
            + self.model.num_hidden_channels,
            :,
            :,
        ]

    @property
    def output_channels(self) -> torch.Tensor:
        """
        Convenience property to access the output channels as a Tensor.

        :returns: BCWH Tensor
        :rtype: torch.Tensor
        """
        return self.output_image[
            :,
            -self.model.num_output_channels :,
            :,
            :,
        ]

    @property
    def output_array(self) -> np.ndarray:
        """
        Convenience property to access the whole output image in the format of
        a numpy array. Brings the entire tensor to CPU on demand, and only at
        the first call.

        :returns: Numpy array in BCWH format
        :rtype: np.ndarray
        """
        if self._output_array is None:
            self._output_array = self.output_image.detach().cpu().numpy()
        return self._output_array

    @property
    def image_channels_np(self) -> np.ndarray:
        """
        Convenience property to access the output image channels in the format of
        a numpy array. Brings the entire tensor to CPU on demand, and only at
        the first call.

        :returns: Numpy array in BCWH format
        :rtype: np.ndarray
        """
        if self._output_array is None:
            self._output_array = self.output_image.detach().cpu().numpy()
        return self._output_array[:, : self.model.num_image_channels, :, :]

    @property
    def hidden_channels_np(self) -> np.ndarray:
        """
        Convenience property to access the hidden image channels in the format of
        a numpy array. Brings the entire tensor to CPU on demand, and only at
        the first call.

        :returns: Numpy array in BCWH format
        :rtype: np.ndarray
        """
        if self._output_array is None:
            self._output_array = self.output_image.detach().cpu().numpy()
        return self._output_array[
            :,
            self.model.num_image_channels : self.model.num_hidden_channels
            + self.model.num_hidden_channels,
            :,
            :,
        ]

    @property
    def output_channels_np(self) -> np.ndarray:
        """
        Convenience property to access the image's output channels in the format of
        a numpy array. Brings the entire tensor to CPU on demand, and only at
        the first call.

        :returns: Numpy array in BCWH format
        :rtype: np.ndarray
        """
        if self._output_array is None:
            self._output_array = self.output_image.detach().cpu().numpy()
        return self._output_array[
            :,
            -self.model.num_output_channels :,
            :,
            :,
        ]



#!/usr/bin/env python3
import pytest

from torch.utils.data import DataLoader

import numpy as np

from ncalab import (
    GrowingNCADataset,
    GrowingNCAModel,
    BasicNCATrainer,
    get_compute_device,
)


def test_growing_training():
    """
    Test if a basic NCA trainer runs through for a few epochs without exception.
    """
    device = get_compute_device("cpu")

    nca = GrowingNCAModel(
        device,
        num_image_channels=4,
        num_hidden_channels=5,
        use_alive_mask=False,
    )

    image = np.zeros((16, 16, 4))
    dataset = GrowingNCADataset(image, nca.num_channels, batch_size=8)
    dataloader_train = DataLoader(dataset, batch_size=8, shuffle=False)

    try:
        trainer = BasicNCATrainer(nca, None, max_epochs=3)
    except Exception as e:
        pytest.fail(str(e))

    try:
        trainer.train(dataloader_train, save_every=100)
    except Exception as e:
        pytest.fail(str(e))



import torch  # type: ignore[import-untyped]
import torch.nn as nn  # type: ignore[import-untyped]
import torch.nn.functional as F  # type: ignore[import-untyped]


class DiceScore(nn.Module):
    """
    Pytorch Module that computes the Dice overlap score between two images.
    """

    def __init__(self):
        super(DiceScore, self).__init__()

    def forward(
        self, x: torch.Tensor, y: torch.Tensor, smooth: float = 1.0
    ) -> torch.Tensor:
        """
        :param x: Reference Input
        :type x: torch.Tensor
        :param y: Other Input
        :type y: torch.Tensor
        :param smooth: Smooting factor, defaults to 1.0
        :type smooth: float

        :returns: Dice score
        :rtype: torch.Tensor
        """
        x = torch.sigmoid(x)
        x = torch.flatten(x)
        y = torch.flatten(y)

        intersection = (x * y).sum()
        dice_score = (2.0 * intersection + smooth) / (x.sum() + y.sum() + smooth)
        return dice_score


class DiceBCELoss(nn.Module):
    """
    Combination of Dice and BCE Loss between two images.
    """

    def __init__(self):
        super(DiceBCELoss, self).__init__()
        self.dicescore = DiceScore()

    def forward(
        self, x: torch.Tensor, y: torch.Tensor, smooth: float = 1.0
    ) -> torch.Tensor:
        """
        :param x: Reference Input
        :type x: torch.Tensor
        :param y: Other Input
        :type y: torch.Tensor
        :param smooth: Smooting factor, defaults to 1.0
        :type smooth: float

        :returns: Dice score
        :rtype: torch.Tensor
        """
        x = torch.sigmoid(x)
        x = torch.flatten(x)
        y = torch.flatten(y)

        dice_loss = 1 - self.dicescore(x, y, smooth)
        BCE = F.binary_cross_entropy(x, y, reduction="mean")
        Dice_BCE = BCE + dice_loss
        return Dice_BCE



import torch

from ncalab import get_compute_device, pad_input, GrowingNCAModel, fix_random_seed


def test_pad_input():
    device = get_compute_device()
    image_channels = 4
    hidden_channels = 12
    batch_size = 8
    W = 32
    H = 32
    total_channels = image_channels + hidden_channels
    nca = GrowingNCAModel(device, 4, 12)
    X = torch.zeros((batch_size, image_channels, W, H))

    X_padded_zero = pad_input(X, nca, noise=False)
    assert X_padded_zero.shape == (batch_size, total_channels, W, H)
    assert torch.all(X_padded_zero[:, image_channels:, :, :] == 0)
    X_padded_noise = pad_input(X, nca, noise=True)
    assert X_padded_noise.shape == (batch_size, total_channels, W, H)
    assert (
        torch.std(
            X_padded_noise[:, image_channels:, :, :], dim=None, keepdim=False
        ).float()
        != 0
    )


def test_fix_random_seed():
    fix_random_seed()
    x_torch_before = torch.rand(10)
    fix_random_seed()
    x_torch_after = torch.rand(10)
    assert torch.all(x_torch_before == x_torch_after)



from pathlib import Path

"""
Current file directory.
"""
__CURRENT_PATH = Path(__file__).resolve().parent

"""
Project root directory.
"""
ROOT_PATH = (__CURRENT_PATH / "..").absolute()

"""
Directory in which training weights are stored.
"""
WEIGHTS_PATH = ROOT_PATH / "weights"

"""
Example task directory.
"""
TASK_PATH = ROOT_PATH / "tasks"

WEIGHTS_PATH.mkdir(exist_ok=True, parents=True)
TASK_PATH.mkdir(exist_ok=True, parents=True)



#!/usr/bin/env python3
import pytest

import albumentations as A  # type: ignore[import-untyped]
from albumentations.pytorch import ToTensorV2  # type: ignore[import-untyped]

from torch.utils.data import Dataset, DataLoader

import numpy as np

from ncalab import (
    DepthNCAModel,
    BasicNCATrainer,
    get_compute_device,
)


T = A.Compose(
    [
        ToTensorV2(),
    ]
)


class DummyDepthDataset(Dataset):
    def __init__(self, transform=None) -> None:
        super().__init__()
        self.images = np.ones((16, 32, 32, 3), dtype=np.float32)
        self.masks = np.ones((16, 32, 32), dtype=np.float32)
        self.transform = transform

    def __len__(self):
        return self.images.shape[0]

    def __getitem__(self, index):
        sample = {"image": self.images[index], "mask": self.masks[index]}
        if self.transform is not None:
            sample = self.transform(**sample)
        return sample["image"], sample["mask"]


def test_depth_training():
    """
    Test if a basic NCA trainer runs through for a few epochs without exception.
    """
    device = get_compute_device("cpu")

    nca = DepthNCAModel(
        device,
        num_image_channels=3,
        num_hidden_channels=5,
        pad_noise=False,
    )

    dataset = DummyDepthDataset(transform=T)
    dataloader_train = DataLoader(dataset, batch_size=8, shuffle=False)

    try:
        trainer = BasicNCATrainer(nca, None, max_epochs=3)
    except Exception as e:
        pytest.fail(str(e))

    try:
        trainer.train(dataloader_train, save_every=100)
    except Exception as e:
        pytest.fail(str(e))


def test_depth_training_with_validation():
    """
    Test if a basic NCA trainer runs through for a few epochs without exception.
    """
    device = get_compute_device("cpu")

    nca = DepthNCAModel(
        device,
        num_image_channels=3,
        num_hidden_channels=5,
        pad_noise=False,
    )

    dataset = DummyDepthDataset(transform=T)
    dataloader_train = DataLoader(dataset, batch_size=8, shuffle=False)
    dataloader_val = DataLoader(dataset, batch_size=8, shuffle=False)

    try:
        trainer = BasicNCATrainer(nca, None, max_epochs=3)
    except Exception as e:
        pytest.fail(str(e))

    try:
        trainer.train(dataloader_train, dataloader_val, save_every=100)
    except Exception as e:
        pytest.fail(str(e))



import logging

import torch


class AutoStepper:
    """
    Helps selecting number of timesteps based on NCA activity.
    """

    def __init__(
        self,
        min_steps: int = 10,
        max_steps: int = 100,
        plateau: int = 5,
        verbose: bool = False,
        threshold: float = 1e-2,
    ):
        """
        :param min_steps: Minimum number of timesteps to always execute, defaults to 10.
        :type min_steps: int, optional
        :param max_steps: Terminate after maximum number of steps, defaults to 100.
        :type max_steps: int, optional
        :param plateau: Number of steps that is considered a plateau, defaults to 5.
        :type plateau: int
        :param verbose: Whether to log interruption to stdout, defaults to False.
        :type verbose: bool
        :param threshold: Score threshold, defaults to 1e-2.
        :type threshold: float
        """
        assert min_steps >= 1
        assert plateau >= 1
        assert max_steps > min_steps
        self.min_steps = min_steps
        self.max_steps = max_steps
        self.plateau = plateau
        self.verbose = verbose
        self.threshold = threshold
        self.cooldown = 0
        # invariant: auto_min_steps > 0, so both of these will be defined when used
        self.hidden_i: torch.Tensor | None = None
        self.hidden_i_1: torch.Tensor | None = None

    def score(self) -> torch.Tensor:
        """
        Calculates activity score.

        Method check() uses this score to determine if the NCA is inactive.

        :return: Activity score estimate.
        :rtype: torch.Tensor
        """
        assert self.hidden_i is not None
        assert self.hidden_i_1 is not None
        # normalized absolute difference between two hidden states
        return (self.hidden_i - self.hidden_i_1).abs().sum() / torch.numel(
            self.hidden_i
        )

    def check(self, step: int) -> bool:
        """
        Checks whether to interrupt inference after the current step.

        :param step: Current NCA inference step.
        :type step: int

        :return: Whether to interrupt inference after the current step.
        :rtype: bool
        """
        with torch.no_grad():
            if step < self.min_steps:
                return False
            if step >= self.max_steps:
                return True
            if self.hidden_i is None or self.hidden_i_1 is None:
                return False
            if self.score() >= self.threshold:
                self.cooldown = 0
            else:
                self.cooldown += 1
            if self.cooldown >= self.plateau:
                if self.verbose:
                    logging.info(f"Breaking after {step} steps.")
                return True
            return False



from __future__ import annotations
import os
import random
from typing import Any

import numpy as np

import torch  # type: ignore[import-untyped]
import torch.nn.functional as F  # type: ignore[import-untyped]


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .models import BasicNCAModel


def get_compute_device(device: str = "cuda:0") -> torch.device:
    """
    Obtain a pytorch compute device handle based on input string.
    If user tries to get a CUDA device, but none is available,
    defaults to CPU.

    :param device: Device string, defaults to "cuda:0".
    :type device: str

    :returns: Pytorch compute device.
    :rtype: torch.device
    """
    if device == "cpu":
        return torch.device("cpu")
    d = torch.device(device if torch.cuda.is_available() else "cpu")
    return d


def pad_input(
    x: torch.Tensor,
    nca: "BasicNCAModel",
    noise: bool = True,
    mean: float = 0.5,
    std: float = 0.225,
) -> torch.Tensor:
    """
    Pads the BCWH input tensor along its channel dimension to match the expected number of
    channels required by the NCA model. Pads with either Gaussian noise (parameterized by
    mean and std) or zeros, depending on the "noise" parameter.

    :param x: Input image tensor, BCWH.
    :type x: torch.Tensor
    :param nca: NCA model definition.
    :type nca: ncalab.BasicNCAModel
    :param noise: Whether to pad with noise. Otherwise zeros, defaults to True.
    :type noise: bool, optional
    :param mean: Mean (mu) of Gaussian noise distribution, defaults to 0.5.
    :type mean: float, optional
    :param std: Standard deviation (sigma) of Gaussian noise distribution, defaults to 0.225.
    :type std: float, optional

    :returns: Input tensor, BCWH, padded along the channel dimension.
    :rtype: torch.Tensor
    """
    if x.shape[1] < nca.num_channels:
        x = F.pad(
            x, (0, 0, 0, 0, 0, nca.num_channels - x.shape[1], 0, 0), mode="constant"
        )
        if noise:
            x[
                :,
                nca.num_image_channels : nca.num_image_channels
                + nca.num_hidden_channels,
                :,
                :,
            ] = torch.normal(
                mean,
                std,
                size=(x.shape[0], nca.num_hidden_channels, x.shape[2], x.shape[3]),
            )
    return x


def print_NCALab_banner():
    """
    Show NCALab banner on terminal.
    """
    banner = """
 _   _  _____          _           _
| \\ | |/ ____|   /\\   | |         | |
|  \\| | |       /  \\  | |     __ _| |__
| . ` | |      / /\\ \\ | |    / _` | '_ \\
| |\\  | |____ / ____ \\| |___| (_| | |_) |
|_| \\_|\\_____/_/    \\_\\______\\__,_|_.__/
-----------------------------------------
    Developed at MECLab - TU Darmstadt
-----------------------------------------
    """
    print(banner)


def print_mascot(message: str):
    """
    Show help text in a speech bubble.

    :param message: Message to display.
    :type message: str
    """
    if not message:
        return
    w = max([len(L) for L in message.splitlines()])
    print("  " + "-" * w)
    for L in message.splitlines():
        print(f"| {L}" + " " * (w - len(L)) + " |")
    print("  " + "=" * w)
    print(" " * w + "   \\")
    print(" " * w + "    \\")

    try:
        print(" " * (w + 3) + "\N{MICROSCOPE}\N{RAT}")
    except UnicodeEncodeError:
        print(" " * (w + 5) + ":3")


"""
Default random seed to use within this project.
"""
DEFAULT_RANDOM_SEED = 1337


def fix_random_seed(seed: int = DEFAULT_RANDOM_SEED):
    """
    Fixes the random seed for all pseudo-random number generators,
    including Python-native, Numpy and Pytorch.

    :param seed: Random seed, defaults to DEFAULT_RANDOM_SEED.
    :type seed: int, optional
    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def unwrap(x: Any):
    """
    Panics if x is None, otherwise returns x.

    This is a useful shorthand for cases such as ``x = unwrap(some_object).do_something()``
    in which we are 99% certain that some_object is not None and want to avoid a mypy complaint.

    :param x: Any kind of object.
    :type x: Any
    :raises RuntimeError: If x is None.
    :return: Just passes through the input x if it is not None.
    """
    if x is None:
        raise RuntimeError("unwrap() failed: Expected return other than None.")
    return x
